#include <fftw3.h>
#include <stdio.h>
#include <math.h>

#include "freq2.hh"

double fft(signed short *data, unsigned length)
{
	double *In = new double[length];
	double *Out = new double[length];
	
	{FILE *wfp = fopen("fftw.wisdom", "rb");
	if(wfp){fftw_import_wisdom_from_file(wfp);
	fclose(wfp);}}

	fftw_plan plan = fftw_plan_r2r_1d
		(length,
		 &In[0],
		 &Out[0],
		 FFTW_R2HC,
		 FFTW_ESTIMATE);

    for(unsigned a=0; a<length; ++a)
	{
		In[a] = 0.5 + data[a] / 65536.0;
		Out[a] = 0;
	}

	fftw_execute(plan);
	
	unsigned puoli = length/2 + 1;
	double *Power = new double[puoli];
	
	Power[0] = fabs(Out[0]);
	
	for(unsigned a=1; a<puoli; ++a)
		Power[a] = sqrt((Out[a] * Out[a]) + (Out[length-a] * Out[length-a]));

	if(!(length%2))
		Power[length/2] = sqrt(Out[length/2] * Out[length/2]);
	
	double biggestspike=0, biggestspikepos=0;
	unsigned spikestart=1;
	unsigned spiketop=0;
	double lastval=0;
	bool ascending=true;
	for(unsigned a=1; a<puoli; ++a)
	{
		const double val = Power[a];
/**/
		fprintf(stderr, "[%u]: power=%g, start=%u,lastval=%g,ascending=%s\n",
				a,val,spikestart,lastval, ascending?"true":"false");
/**/
		if(ascending)
		{
			if(val >= lastval) { lastval = val; continue; }
			// lytyi keskikohta
			spiketop = a-1;
			ascending = false;
			lastval = val;
			continue;
		}
		if(val >= lastval)
		{
			// nousee taas
			ascending = true;

			double leftside = Power[spiketop-1];
			double rightside = Power[spiketop+1];
			double spikepower = Power[spiketop];
			
			double spikepos = spiketop;
			if(rightside != leftside)
			    spikepos += (rightside-leftside) / (rightside+leftside);
/**/
			fprintf(stderr,
					"Power %11.6f spike at %11.4f (%g - %g - %g)\n",
					spikepower, spikepos,
					leftside, Power[spiketop], rightside);
/**/
			if(spikepower > biggestspike)
			{
				biggestspike = spikepower;
				biggestspikepos = spikepos;
			}
			
			spikestart = a-1;
			continue;
		}
		lastval = val;
	}
	
	{FILE *wfp = fopen("fftw.wisdom", "wb");
	if(wfp){fftw_export_wisdom_to_file(wfp);
	fclose(wfp);}}
	
	fftw_destroy_plan(plan);
	
	delete[] Power;
	
	delete[] In;
	delete[] Out;
	
	double wavelen = 2 * length / biggestspikepos;
	
	fprintf(stderr,
		"Length %u: spike %g at %g (wavelen = %g?)\n",
		length, biggestspike, biggestspikepos, wavelen);
	return wavelen;
}

#define MagickPI  3.14159265358979323846264338327950288419716939937510
static double Sinc(double x)
{
  if(x == 0.0)return 1.0;
  x *= MagickPI;
  return sin(x) / x;
}
static double Lanczos(double x)
{
  if(x < 0) x = -x;
  if(x >= 3.0)return 0.0;
  return(Sinc(x)*Sinc(x/3.0));
}

double fft2(signed short *data, unsigned length, double upsamplefactor)
{
	if(upsamplefactor == 1.0)
		return fft(data, length);
	
	const double factor = 1.0/upsamplefactor;
	unsigned resultsize = (unsigned)(length / factor);
	signed short *result = new signed short[resultsize];
	const double support = 3.0 * factor;
	double *weight = new double[(int)(support*2+3)];
	for(unsigned d=0; d<resultsize; ++d)
	{
		double center = d * factor;
		int start = (int)(0.5 + center - support);
		int end   = (int)(0.5 + center + support);
		
		if(start < 0)start = 0;
		if(end >= (int)length)end = length-1;
		
		int n=0;
		{ double density=0;
		for(int i=start; i<end; ++i)
		{
			double distance = i - center + 0.5;
			density += weight[n++] = Lanczos(distance/factor)/factor;
		}
		if(density != 0)
			for(int i=0; i<n; ++i)
				weight[i] /= density; }
		double value=0;
		for(int i=0,p=start; i<n; ++i,++p)
			value += weight[i] * data[p];
		
		if(value < -32768)value = -32768;
		if(value > 32767)value = 32767;
		result[d] = (int)value;
	}
	delete[] weight;
	
	double wavelen = fft(result, resultsize);
	delete[] result;
	return wavelen * factor;
}

#include "reader.hh"

#include <map>
#include <vector>
using std::map;
using std::vector;
using std::make_pair;

double GetWaveLen(unsigned smpid, unsigned rate)
{
	static map<unsigned, double> wavelenmap;
	
	map<unsigned, double>::const_iterator i = wavelenmap.find(smpid);
	if(i != wavelenmap.end()) return i->second;
	
    fprintf(stderr, "Analyzing sample %X ", smpid);
    
    if(Samples.find(smpid) == Samples.end())
    {
    	fprintf(stderr, "- not found!\n");
    	return 1;
    }
    
    const Sample &smp = Samples[smpid];
    vector<short> data;
    
    const short *sampledata = reinterpret_cast<const short *>(smp.data.data());
    unsigned length = smp.data.size() / 2;
    data.insert(data.end(), sampledata, sampledata + length);
    
    if(smp.looping && smp.loopend > smp.loopbeg)
    {
    	data.clear();
        unsigned singlelooplen = smp.loopend - smp.loopbeg;
        unsigned looplen = 0;
        unsigned maxlooplen = data.size() * 3;
        if(maxlooplen > 131000)maxlooplen = 131000;
        if(maxlooplen < rate*2)maxlooplen = rate*2;
        while(looplen < maxlooplen)
        {
            data.insert(data.end(),
                        sampledata + smp.loopbeg,
                        sampledata + smp.loopend);
            looplen += singlelooplen;
        }
    }

    {
    char Buf[512]; sprintf(Buf, "s-%X.raw", smpid);
    FILE *fp = fopen(Buf, "wb");
    fwrite(&data[0], data.size(), sizeof(data[0]), fp);
    fclose(fp); }
            
    fprintf(stderr, "(%u+%u samples)... ", length, data.size()-length);
    
    double wavelen = fft2(&data[0], data.size(), 1.7);
    fprintf(stderr, " got wavelen=%g\n", wavelen);
    
    wavelenmap.insert(make_pair(smpid, wavelen));
    return wavelen;
}
